-- spell names and IDs
local vf = GetSpellInfo(194249)
local stm = GetSpellInfo(193223)
local pi = GetSpellInfo(10060)
local sa = 148859
local swp = 589
local mb = 8092
local vb = 205448
local swd = 32379
local swv = 205351
local sc = 205385

-- initialize frame related stuff
local mainFrame = CreateFrame("Frame", nil, UIParent)
local bgFrame = CreateFrame("Frame", "InsanityBGFrame", mainFrame)
local barFrame = CreateFrame("Frame", "InsanityBarFrame", bgFrame)
local drainFrame = CreateFrame("Frame", "InsanityDrainFrame", barFrame)
local border = 3
local barWidth = 500
local barHeight = 20

local function AddTextToFrame(frame)
	frame.text = frame:CreateFontString(nil, "OVERLAY", "GameFontNormal")
	frame.text:SetText("0.0")
	frame.text:SetPoint("RIGHT", frame, -border, 0)
	frame.text:SetTextColor(1, 1, 1)
end

-- build background frame
bgFrame:SetPoint("CENTER", UIParent)
bgFrame:SetSize(barWidth+2*border, barHeight+2*border)
bgFrame.tex = bgFrame:CreateTexture("InsanityBGTex")
bgFrame.tex:SetAllPoints(bgFrame)
bgFrame.tex:SetColorTexture(0.1, 0.1, 0.1, 0.5)
-- make moveable by mouse
bgFrame:SetMovable(true)
bgFrame:EnableMouse(true)
bgFrame:RegisterForDrag("LeftButton")
bgFrame:SetScript("OnDragStart", bgFrame.StartMoving)
bgFrame:SetScript("OnDragStop", bgFrame.StopMovingOrSizing)
bgFrame:SetUserPlaced(true)

-- build bar frame
barFrame:SetPoint("LEFT", bgFrame, "LEFT", border, 0)
barFrame.tex = barFrame:CreateTexture("InsanityBarTex")
barFrame.tex:SetAllPoints(barFrame)
barFrame.tex:SetColorTexture(0.4, 0.2, 1.0, 0.7)
-- add text frame
AddTextToFrame(barFrame)

-- build drain frame
drainFrame:SetPoint("LEFT", bgFrame, "LEFT", border, 0)
drainFrame.tex = drainFrame:CreateTexture("InsanityDrainTex")
drainFrame.tex:SetAllPoints(drainFrame)
drainFrame.tex:SetColorTexture(0.5, 0.0, 0.0, 0.5)
drainFrame:SetFrameStrata("DIALOG")

-- other stuff
local spells = {}
local extraInsFrames = {}
local swdThresh = 0
local saInFlight = 0


local function UpdateInsanitySpells()
	-- dictionary of spells that grant a good amount of insanity and their corresponding amount
	-- the DoTs are not included here, but spells that could prevent dropping out of shadow form
	
	-- check for STM multiplier
	local stmMult = 1
	local _, _, _, selected = GetTalentInfo(7, 3, 1)
	if selected then
		stmMult = 3
	end
	
	-- check for RoS multiplier
	local rosMult = 1
	swdThresh = 0.2
	local _, _, _, selected = GetTalentInfo(4, 2, 1)
	if selected then
		rosMult = 3
		swdThresh = 0.35
	end
	
	spells = {}
	spells[mb] = 12 -- mindblast
	spells[vb] = 16 -- void bolt
	spells[swd] = 10*rosMult -- shadow word: death
	
	-- check for talents selected that grant insanity
	local _, _, _, selected = GetTalentInfo(1, 3, 1)
	if selected then
		spells[swv] = 25 -- shadow word: void
	end
	local _, _, _, selected = GetTalentInfo(5, 2, 1)
	if selected then
		spells[sa] = 4 -- shadowy appariations
	end
	local _, _, _, selected = GetTalentInfo(6, 2, 1)
	if selected then
		spells[sc] = 15 -- shadow crash
	end
	
	-- build extra insanity frames
	for spellId, extraIns in pairs(spells) do
		if not extraInsFrames[spellId] then
			-- make new frame because not yet in list
			local f = CreateFrame("Frame", "ExtraInsanityBarFrame"..spellId, barFrame)
			f:SetPoint("LEFT", barFrame, "RIGHT", 0, 0)
			f.tex = f:CreateTexture("ExtraInsanityBarTex"..spellId)
			f.tex:SetAllPoints(f)
			f.tex:SetColorTexture(0.7, 0.9, 1.0, 0.7)
			-- add text frame
			AddTextToFrame(f)
			
			-- also make icon
			local g = CreateFrame("Frame", "ExtraInsanityIconFrame"..spellId, f)
			g:SetPoint("LEFT", f, "LEFT", 0, 0)
			g:SetSize(barHeight, barHeight)
			g.tex = g:CreateTexture("ExtraInsanityIconTex"..spellId)
			g.tex:SetAllPoints(g)
			local texFile = GetSpellTexture(spellId)
			g.tex:SetTexture(texFile)
			
			-- link and hide
			f.icon = g
			f:Hide()
			-- add to list
			extraInsFrames[spellId] = f
		end
	end
end

local function InsDrain(stacks)
	-- insanity drain per second based on stacks (from voidform)
	return 8+(stacks-1)/2
end

local function TimeLeft(curIns, stacks)
	-- returns the remaining time until insanity would be completely drained with stacks of voidform
	-- calculation: F(x) = INTEGRAL of InsDrain(x), y = curIns-F(x-stacks)+F(stacks), get x for y = 0
	return math.sqrt(4*curIns+stacks*stacks+30*stacks+225)-stacks-15
end

local function UpdateBar()
	-- the main routine showing current insanity and extra insanity if spells are ready
	-- unfortunately "UnitPower" seems not be updated that often
	-- even when this function is executed using the OnUpdate script
	
	-- current insanity level
	local curIns = UnitPower("player")
	local curWidth = barWidth*curIns/100
	barFrame:SetSize(curWidth, barHeight)
	
	-- current drain amount
	local vfUp, _, _, counts = UnitBuff("player", vf)
	local stacks, timeTilDrained = 0, 0
	if vfUp then
		stacks = min(counts, 100)
		drainFrame:SetSize(barWidth*InsDrain(stacks)/100, barHeight)
		-- get seconds until insanity would be drained to 0 for current stacks
		timeTilDrained = TimeLeft(curIns, stacks)
		barFrame.text:SetText(string.format("%.1f", timeTilDrained))
	end
	
	-- insanity multiplier to all spells
	local globInsMult = 1
	-- surrender to madness
	local stmUp = UnitBuff("player", stm)
	if stmUp then
		globInsMult = globInsMult*3
	end
	-- power infusion
	local piUp = UnitBuff("player", pi)
	if piUp then
		globInsMult = globInsMult*1.25
	end
	
	-- append extra insanity
	local prevFrame = barFrame
	for spellId, extraIns in pairs(spells) do
		-- check if spell is off cooldown
		local f = extraInsFrames[spellId]
		local start, duration = GetSpellCooldown(spellId)
		local time = GetTime()
		local cd = start + duration - time
		local cdReady = not (cd > 0 and duration > 2)
		
		-- condition for various cases
		cond = false
		if spellId == swd then
			-- special case for shadow word: death
			cond = cdReady and UnitExists("target") and UnitHealth("target")/UnitHealthMax("target") <= swdThresh
		elseif spellId == sa then
			-- special case for SA's
			cond = saInFlight > 0
			extraIns = extraIns*saInFlight
		else
			-- normal castable spells
			cond = cdReady
		end
		
		-- show extra insanity when condition is met
		if cond then
			f:Show()
			f:SetPoint("LEFT", prevFrame, "RIGHT", 0, 0)
			f:SetSize(extraIns*globInsMult*barWidth/100, barHeight)
			prevFrame = f
			-- show extra time
			if vfUp then
				-- get gained seconds
				local extraTimeTilDrained = TimeLeft(extraIns*globInsMult+curIns, stacks)
				f.text:SetText(string.format("+%.1f", extraTimeTilDrained-timeTilDrained))
			end
		else
			f:Hide()
		end
	end
end

-- watchdog to reset SA counter in case something goes wrong
local saWatchdog = 0
mainFrame:SetScript("OnUpdate", function(self, elapsed)
	if (saWatchdog > 0) then
		saWatchdog = saWatchdog-elapsed
	else
		saInFlight = 0
	end
end)


-- events
mainFrame:RegisterEvent("PLAYER_ENTERING_WORLD")
mainFrame:RegisterEvent("PLAYER_TALENT_UPDATE")
mainFrame:RegisterEvent("UNIT_AURA")
mainFrame:SetScript("OnEvent", function(self, event, ...)
	-- look for talents selected
	if event == "PLAYER_ENTERING_WORLD" or event == "PLAYER_TALENT_UPDATE" then
		UpdateInsanitySpells()
	end
	
	-- look if voidform is up
	if event == "UNIT_AURA" then
		-- activate insanity bar while buff is up
		local buffUp = UnitBuff("player", vf)
		if buffUp then
			bgFrame:Show()
			bgFrame:RegisterEvent("UNIT_POWER_FREQUENT")
		else
			bgFrame:Hide()
			bgFrame:UnregisterEvent("UNIT_POWER_FREQUENT")
		end
	end
end)

bgFrame:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
bgFrame:SetScript("OnEvent", function(self, event, ...)
	-- this is fired when the main resource of a unit changes
	if event == "UNIT_POWER_FREQUENT" then
		local unit, powerType = ...
		if unit == "player" and powerType == "INSANITY" then
			UpdateBar()
		end
	end
	-- this is fired for every combat action - use with caution
	if event == "COMBAT_LOG_EVENT_UNFILTERED" then
		-- we need this to track how many SA's are on the way (ideally)
		local _, spellEvent, _, _, sourceUnit, _, _, _, _, _, _, spellId, _, _, _, _, _, _, _, _, critical = ...
		-- only track stuff from us
		if sourceUnit == "Azial" then
			-- track shadowy appariations impact
			if spellId == sa and spellEvent == "SPELL_DAMAGE" then
				saInFlight = saInFlight-1
			end
			-- track crits from sw:p
			if spellId == swp and spellEvent == "SPELL_PERIODIC_DAMAGE" and critical then
				saInFlight = saInFlight+1
				saWatchdog = 10 -- reset watchdog to x seconds
			end
		end
	end
end)